@@ -29,14 +29,28 @@
 #include <ctype.h>
 #include "../neogb/libneogb.h"
 
+#include <flint/nmod_poly.h>
 #define MODP(a,b)                               \
 	(a) % (b)
+
+
+typedef struct
+{
+  mpz_t numer;
+  long k;
+  unsigned int isexact;
+  int sign_left;
+} interval;
+
 typedef len_t nelts_t;
 typedef int32_t nvars_t;
+typedef int64_t bits_t;
+
+
 
 typedef struct{
-  int32_t nvars;
-  int32_t elim;
+  nvars_t nvars;
+  nvars_t elim;
   int32_t ngens;
   int32_t nterms;
   int32_t field_char;
@@ -60,6 +74,96 @@ typedef struct{
   mpz_t **mpz_cfs; /* mpz_t coeffs */
 } data_gens_ff_t;
 
+typedef uint32_t szmat_t;
+
+typedef uint32_t CF_t;
+typedef uint64_t CF_l_t;
+typedef uint32_t mod_t;
+typedef int32_t nvars_t;
+/* typedef __uint128_t CF_L_t; */
+
+/**
+
+M:=Matrix([[0, 1, 0, 0, 0],[ 0, 0, 0, 1, 0],[1, 2, 4, 5, 6],[0, 0, 0, 0, 1],[7, 8, 1, 9, 4]]);
+
+   Une matrice de la forme
+
+   0 1 0 0 0
+   0 0 0 1 0
+   1 2 4 5 6
+   0 0 0 0 1
+   7 8 1 9 4
+
+   est code de la forme
+   ncols  = 5
+   nrows = 2
+   dense_mat -> 1 2 4 5 6 7 8 1 9
+   triv_idx -> 0 1 3 (lgr donnee par ncols-nrows)
+   triv_pos -> 1 3 4
+   dense_idx -> 2 4 (lgr donnee par nrows)
+ **/
+
+typedef struct{
+  CF_t charac;
+  szmat_t ncols; //dimension du quotient
+  szmat_t nrows; //nbre de lignes non triviales
+  CF_t *dense_mat; // matrice nrows lignes et ncols colonnes (elements donnes par lignes)
+  szmat_t *triv_idx; //tableau d'indices des lignes ne contenant que des 0 et un 1
+  szmat_t *triv_pos; //position des 1
+  szmat_t *dense_idx; //position des lignes non triviales (qui constituent donc
+                      //dense_mat)
+  szmat_t *dst; //pour la gestion des lignes "denses" mais avec un bloc de zero a la fin
+} sp_matfglm_t;
+
+#ifndef ALIGNED32
+#define ALIGNED32 __attribute__((aligned(32)))
+#endif
+typedef struct{
+  CF_t *vecinit ALIGNED32; //vecteur initial
+  CF_t *res ALIGNED32; //va contenir les termes de suites qui nous interessent
+  CF_t *vecmult ALIGNED32; //utilise pour la multiplication
+  CF_t *vvec ALIGNED32;
+  CF_l_t *vec_cache ALIGNED32; //useless
+  mp_limb_t *pts;
+} fglm_data_t;
+
+
+typedef struct{
+  nmod_berlekamp_massey_t BMS;
+  nmod_poly_t Z1;
+  nmod_poly_t Z2;
+  nmod_poly_t rZ1;
+  nmod_poly_t rZ2;
+  nmod_poly_t A;
+  nmod_poly_t B;
+  nmod_poly_t V;
+  nmod_poly_t param;
+  nmod_poly_factor_t sqf;
+} fglm_bms_data_t;
+
+typedef struct{
+  CF_t charac;
+  szmat_t ncols; //dimension du sev du quotient
+  szmat_t nrows; //nbre de lignes non triviales
+  szmat_t nzero; //nbre de lignes nulles
+  CF_t *dense_mat; // matrice nrows lignes et ncols colonnes (elements donnes par lignes)
+  szmat_t *triv_idx; //tableau d'indices des lignes ne contenant que des 0 et un 1
+  szmat_t *triv_pos; //position des 1
+  szmat_t *dense_idx; //position des lignes non triviales (qui constituent donc
+                      //dense_mat)
+  szmat_t *zero_idx; //tableau d'indices des lignes ne contenant que des 0
+  szmat_t *dst; //pour la gestion des lignes "denses" mais avec un bloc de zero a la fin
+} sp_matfglmcol_t;
+
+typedef struct{
+  mp_limb_t charac;
+  nvars_t nvars;
+  nmod_poly_t elim;
+  nmod_poly_t denom;
+  nmod_poly_t *coords;
+} param_t;
+
+
 typedef struct{
   mpz_t r0;
   mpz_t r1;
@@ -75,18 +179,18 @@ typedef rrec_data_struct_t rrec_data_t[1];
 
 
 typedef struct{
-  int32_t length;
-  int32_t alloc;
+  deg_t length;
+  deg_t alloc;
   mpz_t *coeffs;
 } mpz_upoly_struct;
 
 typedef mpz_upoly_struct mpz_upoly_t[1];
 
 typedef struct{
-  long nvars;
-  long nsols;
-  long dquot;
-  int dim;
+  nvars_t nvars;
+  deg_t nsols;
+  len_t dquot;
+  deg_t dim;
   mpz_upoly_t elim;
   mpz_upoly_t denom;
   mpz_upoly_t *coords;
@@ -95,7 +199,7 @@ typedef struct{
 typedef mpz_param_struct mpz_param_t[1];
 
 typedef struct{
-  int32_t nb;
+  len_t nb;
   mpz_param_t *params;
 } mpz_param_array_struct;
 typedef mpz_param_array_struct mpz_param_array_t[1];
@@ -144,24 +248,24 @@ typedef struct{
   mpz_t det_num;
   mpz_t det_den;
   mpz_t tmp;
-  int done_trace;
-  int done_det;
-  int check_trace;
-  int check_det;
+  int16_t done_trace;
+  int16_t done_det;
+  int16_t check_trace;
+  int16_t check_det;
 } trace_det_fglm_mat_struct;
 typedef trace_det_fglm_mat_struct trace_det_fglm_mat_t[1];
 
 typedef struct{
   mpz_t val_up;
   mpz_t val_do;
-  long k_up;
-  long k_do;
+  deg_t k_up;
+  deg_t k_do;
   unsigned int isexact;
 } coord_struct;
 typedef coord_struct coord_t[1];
 
 typedef struct{
-  long nvars;
+  nvars_t nvars;
   coord_t *coords;
 } real_point_struct;
 typedef real_point_struct real_point_t[1];
